"""
Our Models
"""

# Django
from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from eveuniverse.models import EveType


def get_sentinel_user():
    """
    get user or create one
    :return:
    """

    return User.objects.get_or_create(username="deleted")[0]


class AaSrpStatus(models.TextChoices):
    """
    Choices for SRP Status
    """

    ACTIVE = "Active", _("Active")
    CLOSED = "Closed", _("Closed")
    COMPLETED = "Completed", _("Completed")


class AaSrpRequestStatus(models.TextChoices):
    """
    Choices for SRP Request Status
    """

    PENDING = "Pending", _("Pending")
    APPROVED = "Approved", _("Approved")
    REJECTED = "Rejected", _("Rejected")


class AaSrpRequestCommentType(models.TextChoices):
    """
    Choices for Comment Types
    """

    COMMENT = "Comment", _("SRP Request Comment")
    REQUEST_INFO = "Request Information", _("SRP Request Additional Information")
    REJECT_REASON = "Reject Reason", _("SRP Reject Reason")


class AaSrp(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:
        verbose_name = "AA-SRP"
        managed = False
        default_permissions = ()
        permissions = (
            # can open the SRP app and submit SRP requests
            ("basic_access", "Can access the AA-SRP module"),
            # can create SRP links
            ("create_srp", "Can create new SRP links"),
            # can manage the complete SRP module
            ("manage_srp", "Can manage SRP"),
            # can manage SRP requests only
            ("manage_srp_requests", "Can manage SRP requests"),
        )


class AaSrpLink(models.Model):
    """
    SRP Link model
    """

    srp_name = models.CharField(max_length=254, default="")
    srp_status = models.CharField(
        max_length=9, choices=AaSrpStatus.choices, default=AaSrpStatus.ACTIVE
    )
    srp_code = models.CharField(max_length=16, default="")
    fleet_commander = models.ForeignKey(
        EveCharacter,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
    )
    fleet_doctrine = models.CharField(max_length=254, default="")
    fleet_time = models.DateTimeField()
    aar_link = models.CharField(max_length=254, blank=True, default="")

    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
        help_text=_("Who created the SRP link?"),
    )

    def __str__(self) -> str:
        return str(self.srp_name)

    @property
    def total_cost(self):
        """
        total cost for this SRP link
        :return:
        """

        return sum(int(r.payout_amount) for r in self.srp_requests.all())

    @property
    def total_requests(self):
        """
        Number of total SRP requests
        :return:
        """

        return self.srp_requests.count()

    @property
    def pending_requests(self):
        """
        Number of pending SRP requests
        :return:
        """

        return self.srp_requests.filter(
            request_status=AaSrpRequestStatus.PENDING
        ).count()

    @property
    def approved_requests(self):
        """
        Number of approved SRP requests
        :return:
        """

        return self.srp_requests.filter(
            request_status=AaSrpRequestStatus.APPROVED
        ).count()

    @property
    def rejected_requests(self):
        """
        Number of rejected SRP requests
        :return:
        """

        return self.srp_requests.filter(
            request_status=AaSrpRequestStatus.REJECTED
        ).count()

    @property
    def requests(self):
        """
        All SRP requests
        :return:
        """

        return self.srp_requests.all()

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("SRP Link")
        verbose_name_plural = _("SRP Links")


class AaSrpRequest(models.Model):
    """
    SRP Request model
    """

    request_code = models.CharField(max_length=254, default="")
    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
        help_text=_("Who created the SRP link?"),
    )
    character = models.ForeignKey(EveCharacter, null=True, on_delete=models.SET_NULL)
    ship_name = models.CharField(max_length=254, default="")
    ship = models.ForeignKey(
        EveType,
        related_name="srp_requests",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
    )
    killboard_link = models.CharField(max_length=254, default="")
    additional_info = models.TextField(blank=True, default="")
    request_status = models.CharField(
        max_length=8,
        choices=AaSrpRequestStatus.choices,
        default=AaSrpRequestStatus.PENDING,
    )
    payout_amount = models.BigIntegerField(default=0)
    srp_link = models.ForeignKey(
        AaSrpLink, related_name="srp_requests", on_delete=models.CASCADE
    )
    loss_amount = models.BigIntegerField(default=0)
    post_time = models.DateTimeField(default=timezone.now)
    reject_info = models.TextField(blank=True, default="")

    def __str__(self):
        character_name = self.character.character_name
        user_name = self.creator.profile.main_character.character_name
        ship = self.ship.name
        request_code = self.request_code

        return _(
            f"{character_name} ({user_name}) SRP Request for: {ship} ({request_code})"
        )

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("SRP Request")
        verbose_name_plural = _("SRP Requests")


class AaSrpInsurance(models.Model):
    """
    Insurance Model
    """

    srp_request = models.ForeignKey(
        AaSrpRequest, on_delete=models.CASCADE, related_name="insurance"
    )
    insurance_level = models.CharField(max_length=254, default="")
    insurance_cost = models.FloatField()
    insurance_payout = models.FloatField()

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Ship Insurance")
        verbose_name_plural = _("Ship Insurances")


class AaSrpRequestComment(models.Model):
    """
    SRP Request Comments model
    """

    comment = models.TextField(null=True, blank=True)

    comment_type = models.CharField(
        max_length=19,
        choices=AaSrpRequestCommentType.choices,
        default=AaSrpRequestCommentType.COMMENT,
    )

    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
    )

    srp_request = models.ForeignKey(
        AaSrpRequest,
        related_name="srp_request_comments",
        null=True,
        blank=True,
        default=None,
        on_delete=models.CASCADE,
    )

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("SRP Request Comment")
        verbose_name_plural = _("SRP Request Comments")


class AaSrpUserSettings(models.Model):
    """
    User settings
    """

    user = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
    )

    disable_notifications = models.BooleanField(default=False)

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("User Settings")
        verbose_name_plural = _("User Settings")



"""
notifications helper
"""


# Django
from django.contrib.auth.models import User
from django.utils import timezone

# Alliance Auth
from allianceauth.notifications import notify

# AA SRP
from aasrp import __title__
from aasrp.app_settings import aa_discordnotify_active, allianceauth_discordbot_active
from aasrp.constants import DISCORD_EMBED_COLOR_MAP


def send_user_notification(user: User, level: str, title: str, message: str) -> None:
    """
    send notification to user
    this creates a notification in Auth and a PM in Discord when either
    AA-Discordbot or AA Discord Notifications are installed
    :param user:
    :param level:
    :param title:
    :param message:
    """

    notify(user=user, title=title, level=level, message=message)

    # Handle Discord PMs when aa_discordnotify is not active
    # Check if either allianceauth_discordbot or discordproxy are available
    # to send the PM
    if not aa_discordnotify_active():
        try:
            # Third Party
            from discordproxy.client import DiscordClient
            from discordproxy.discord_api_pb2 import Embed
            from discordproxy.exceptions import DiscordProxyException
        except ModuleNotFoundError:
            # discordproxy not available, try allianceauth-discordbot
            if allianceauth_discordbot_active():
                # Third Party
                import aadiscordbot.tasks

                aadiscordbot.tasks.send_direct_message_by_user_id.delay(
                    user.pk, message
                )
        else:
            # discordproxy is available, use it,
            # if the user has a Discord account that is
            if hasattr(user, "discord"):
                client = DiscordClient()

                footer = Embed.Footer(text=__title__)

                embed = Embed(
                    title=str(title),
                    description=message,
                    color=DISCORD_EMBED_COLOR_MAP.get(level, None),
                    timestamp=timezone.now().isoformat(),
                    footer=footer,
                )

                try:
                    client.create_direct_message(
                        user_id=int(user.discord.uid), embed=embed
                    )
                except DiscordProxyException:
                    # something went wrong with discordproxy, fail silently
                    pass


def send_message_to_discord_channel(
    channel_id: int, title: str, message: str, embed: bool = False
) -> None:
    """
    sending a message to a discord channel
    if AA-Discordbot is installed
    :param channel_id:
    :param title:
    :param message:
    :param embed:
    """

    # Check if either allianceauth_discordbot or discordproxy are available
    # to send the channel message
    try:
        # Third Party
        from discordproxy.client import DiscordClient
        from discordproxy.discord_api_pb2 import Embed
        from discordproxy.exceptions import DiscordProxyException
    except ModuleNotFoundError:
        # discordproxy not available, try allianceauth-discordbot
        if allianceauth_discordbot_active():
            # Third Party
            import aadiscordbot.tasks

            message_to_send = f"**{title}**\n\n{message}"

            aadiscordbot.tasks.send_channel_message_by_discord_id.delay(
                channel_id, message_to_send, embed
            )
    else:
        # discordproxy is available, use it
        client = DiscordClient()

        footer = Embed.Footer(text=__title__)

        embed = Embed(
            title=title,
            description=message,
            color=DISCORD_EMBED_COLOR_MAP.get("info", None),
            timestamp=timezone.now().isoformat(),
            footer=footer,
        )

        try:
            client.create_channel_message(channel_id=channel_id, embed=embed)
        except DiscordProxyException:
            # something went wrong with discordproxy, fail silently
            pass



"""
Migrate comments from AaSrpRequest to AaSrpRequestComments
"""

# Django
from django.core.management import BaseCommand

# AA SRP
from aasrp.models import (
    AaSrpRequest,
    AaSrpRequestComment,
    AaSrpRequestCommentType,
    AaSrpRequestStatus,
)


def get_input(text):
    """
    wrapped input to enable tz import
    """

    return input(text)


class Command(BaseCommand):
    """
    Migrate comments from AaSrpRequest to AaSrpRequestComments
    """

    help = "Migrate comments from AaSrpRequest to AaSrpRequestComments"

    def _migrate_reject_comments(self) -> None:
        """
        start migration
        :return:
        """

        srp_requests = AaSrpRequest.objects.all()

        self.stdout.write(
            self.style.WARNING(f"Found {srp_requests.count()} SRP requests.")
        )

        if srp_requests.count() > 0:
            for srp_request in srp_requests:
                reject_reason = srp_request.reject_info
                srp_info = srp_request.additional_info
                request_code = srp_request.request_code

                # migrate additional info
                if srp_info != "":
                    self.stdout.write(
                        f"Updating Request Information for SRP request {request_code}"
                    )

                    # check if there is already a request info
                    # for this request and remove it
                    AaSrpRequestComment.objects.filter(
                        srp_request=srp_request,
                        comment_type=AaSrpRequestCommentType.REQUEST_INFO,
                    ).delete()

                    # write new
                    srp_request_comment = AaSrpRequestComment()
                    srp_request_comment.srp_request = srp_request
                    srp_request_comment.comment = srp_info
                    srp_request_comment.comment_type = (
                        AaSrpRequestCommentType.REQUEST_INFO
                    )
                    srp_request_comment.creator = srp_request.creator
                    srp_request_comment.save()

                    # remove it from AaSrpRequest
                    srp_request.additional_info = ""
                    srp_request.save()

                # migrate reject reason
                if (
                    srp_request.request_status == AaSrpRequestStatus.REJECTED
                    and reject_reason != ""
                ):
                    self.stdout.write(
                        f"Updating Reject Information for SRP request {request_code}"
                    )

                    # check if there is already a reject comment
                    # for this request and remove it
                    AaSrpRequestComment.objects.filter(
                        srp_request=srp_request,
                        comment_type=AaSrpRequestCommentType.REJECT_REASON,
                    ).delete()

                    # write new
                    srp_request_comment = AaSrpRequestComment()
                    srp_request_comment.srp_request = srp_request
                    srp_request_comment.comment = reject_reason
                    srp_request_comment.comment_type = (
                        AaSrpRequestCommentType.REJECT_REASON
                    )
                    srp_request_comment.save()

                    # remove it from AaSrpRequest
                    srp_request.reject_info = ""
                    srp_request.save()

    def handle(self, *args, **options):
        """
        ask before running ...
        :param args:
        :param options:
        """

        self.stdout.write(
            "This command will migrate the reject comments "
            "from AaSrpRequest to AaSrpRequestComments"
        )

        user_input = get_input("Are you sure you want to proceed? (yes/no)?")

        if user_input == "yes":
            self.stdout.write("Starting migration. Please stand by.")
            self._migrate_reject_comments()
            self.stdout.write(self.style.SUCCESS("Update complete!"))
        else:
            self.stdout.write(self.style.WARNING("Aborted."))
